"""
State changes for accounts.

Also, the account is the aggregate root for invoices and charges,
so the creation of those is managed here.

"""
from datetime import date
from typing import Dict, List, Optional, Sequence, Tuple
from uuid import UUID

from django.db import transaction
from moneyed import Money
from structlog import get_logger

from billing.signals import invoice_ready
from . import invoices
from .invoices import PreconditionError
from ..models import (
    Account, CARRIED_FORWARD, CREDIT_REMAINING, Charge, CreditCard,
    EventLog, Invoice, ProductProperty, Transaction, total_amount,
)

logger = get_logger()


def close(account_id: str) -> None:
    """
    Closes the account.

    :param account_id: the account to close
    :return: Nothing
    """
    logger.info('closing-account', account_id=account_id)
    with transaction.atomic():
        account = Account.objects.get(pk=account_id)
        account.close()
        account.save()


def reopen(account_id: str) -> None:
    """
    Reopens the account.

    :param account_id: the account to reopen
    :return: Nothing
    """
    logger.info('reopening-account', account_id=account_id)
    with transaction.atomic():
        account = Account.objects.get(pk=account_id)
        account.reopen()
        account.save()


def create_invoices(account_id: str, due_date: date) -> Sequence[Invoice]:
    """
    Creates the invoices for any due positive charges in the account.
    If there are due positive charges in different currencies, one invoice is created for each currency.

    :param account_id: The account to invoice.
    :param due_date: The due date for any invoice that gets created.
    :return: A possibly-empty list of Invoices.
    """
    invoices = []
    with transaction.atomic():
        due_charges = Charge.objects \
            .uninvoiced(account_id=account_id) \
            .charges()
        total = total_amount(due_charges)
        for amount_due in total.monies():
            if amount_due.amount > 0:
                invoice = Invoice.objects.create(account_id=account_id, due_date=due_date)
                Charge.objects \
                    .uninvoiced(account_id=account_id) \
                    .charges() \
                    .in_currency(currency=amount_due.currency) \
                    .update(invoice=invoice)
                invoices.append(invoice)
    logger.info('created-invoices', account_id=str(account_id), invoice_ids=[i.pk for i in invoices])
    for invoice in invoices:
        invoice_ready.send(sender=create_invoices, invoice=invoice)
    return invoices


def add_charge(account_id: str,
               amount: Money,
               reverses_id: Optional[str] = None,
               product_code: Optional[str] = None,
               product_properties: Optional[Dict[str, str]] = None) -> Charge:
    """
    Add a charge to the account.

    :param account_id: The account on which to add the charge
    :param amount:  The amount of the charge
    :param reverses_id: Set this if this charge reverses another one
    :param product_code: A code identifying the type of product cnarged
    :param product_properties: A dict of hames and values.
    :return: The newly created charge.
    """
    logger.info('adding-charge', account_id=account_id, amount=amount, product_code=product_code,
                product_properties=product_properties)

    with transaction.atomic():
        charge = Charge(account_id=account_id,
                        amount=amount)
        if reverses_id:
            charge.reverses_id = reverses_id
        if product_code:
            charge.product_code = product_code
        charge.full_clean(exclude=['id', 'account'])  # Exclude to avoid unnecessary db queries
        charge.save(force_insert=True)

        if product_properties:
            objs = [ProductProperty(charge=charge, name=k, value=v) for k, v in product_properties.items()]
            for o in objs:
                o.full_clean(exclude=['id', 'charge'])  # Exclude to avoid unnecessary db queries
            ProductProperty.objects.bulk_create(objs)

    return charge


def assign_funds_to_account_pending_invoices(account_id: str) -> Sequence[str]:
    """
    Tries to pay pending account invoices (starting from the oldest) with available funds.
    :param account_id: the account on which to perform the operation
    :return: The ids of the invoices that were paid (possibly empty list).
    """
    logger.info('assign-funds-to-pending-invoices', account_id=str(account_id))

    paid_invoice_ids = []
    for invoice in Invoice.objects.filter(status=Invoice.PENDING, account_id=account_id).order_by('due_date'):
        invoice_was_paid = assign_funds_to_invoice(invoice.pk)
        if invoice_was_paid:
            paid_invoice_ids.append(invoice.id)
        else:
            break  # Bail even though there may be funds in another currency to pay more recent invoices.
    logger.info('assign-funds-to-pending-invoices.end', account_id=str(account_id),
                paid_invoice_count=len(paid_invoice_ids))
    return paid_invoice_ids


def assign_funds_to_invoice(invoice_id: str) -> bool:
    """
    Uses the available funds on the account (credits and payments) to pay the given invoice.
    :param invoice_id: The id of the invoice.
    :return: True if the invoice status is paid.

    A lot of side effects may occur in the database:
    - Funds (either payments or credits) may get assigned to the invoice.
    - The invoice status may change.
    - Credits entities may be created.
    """

    logger.info('assign-funds-to-invoice', invoice_id=invoice_id)
    invoice = Invoice.objects.get(pk=invoice_id)
    account_id = invoice.account_id

    #
    # Precondition. Don't touch invoices that are not PENDING
    #
    if invoice.status != Invoice.PENDING:
        logger.info('assign-funds-to-invoice.status-is-not-pending', invoice_id=invoice_id)
        return False

    #
    # Precondition: Only handle invoices in a single currency
    #
    invoice_due_monies = invoice.due().monies()
    if len(invoice_due_monies) != 1:
        logger.info('assign-funds-to-invoice.more-than-one-currency', invoice_id=invoice_id)
        return False
    invoice_due_amount = invoice_due_monies[0].amount
    invoice_due_currency = invoice_due_monies[0].currency

    #
    # 1. Collect funds as long as long as we need them
    #
    if invoice_due_amount > 0:

        payments = Transaction.successful \
            .payments() \
            .uninvoiced(account_id=account_id) \
            .in_currency(invoice_due_currency) \
            .order_by('created')

        credits = Charge.objects \
            .credits() \
            .uninvoiced(account_id=account_id) \
            .in_currency(invoice_due_currency) \
            .order_by('created')

        funds = list(credits) + list(payments)
        for fund in funds:
            contributed_amount = abs(fund.amount.amount)  # 'abs' because credits have a negative value
            logger.info('assign-funds-to-invoice.assigning-fund',
                        invoice_id=invoice_id,
                        fund_type=type(fund).__name__,
                        fund_id=str(fund.pk),
                        contributed_amount=contributed_amount)
            fund.invoice_id = invoice_id
            fund.save()
            invoice_due_amount -= contributed_amount
            if invoice_due_amount <= 0:
                break

    #
    # 2. Mark invoice paid if nothing is due.
    #
    if invoice_due_amount <= 0:
        logger.info('assign-funds-to-invoice.mark-paid', invoice_id=invoice_id, invoice_due_amount=invoice_due_amount)
        invoice.status = Invoice.PAID
        invoice.save()

    #
    # 3. Carry forward any overpaid money.
    #
    if invoice_due_amount < 0:
        overpayment = Money(abs(invoice_due_amount), invoice_due_currency)
        logger.info('assign-funds-to-invoice.handling-overpayment',
                    invoice_id=invoice_id,
                    overpayment=overpayment)
        with transaction.atomic():
            Charge.objects.create(account_id=account_id, amount=overpayment, product_code=CARRIED_FORWARD,
                                  invoice_id=invoice_id)
            Charge.objects.create(account_id=account_id, amount=-overpayment, product_code=CREDIT_REMAINING)

    return invoice.status == Invoice.PAID


def get_accounts_which_delinquent_status_has_to_change(
    account_ids: List[UUID]
) -> Tuple[List[UUID], List[UUID]]:
    accounts = Account.objects.filter(id__in=account_ids)
    new_delinquent_account_ids = []
    new_compliant_account_ids = []
    for account in accounts:
        reasons = get_reasons_account_is_violating_delinquent_criteria(account.id)
        if reasons and not account.delinquent:
            new_delinquent_account_ids.append(account.id)
        if not reasons and account.delinquent:
            new_compliant_account_ids.append(account.id)

    return new_delinquent_account_ids, new_compliant_account_ids


def get_reasons_account_is_violating_delinquent_criteria(
    account_id: UUID
) -> List[str]:
    reasons = []
    account = Account.objects.get(id=account_id)
    if account.invoices.filter(status=Invoice.PENDING).count() > 0:
        reasons.append('Account has pending invoices')

    if not CreditCard.objects.filter(account=account).valid().exists():
        reasons.append('Account has not any valid credit card registered')

    return reasons


def mark_account_as_delinquent(account_id: UUID, reason: str):
    account = Account.objects.get(id=account_id)
    if not account.delinquent:
        logger.info('mark-account-as-delinquent', account_id=account_id, reason=reason)
        account.delinquent = True
        account.save()
        EventLog.objects.create(
            account_id=account_id,
            type=EventLog.NEW_DELINQUENT,
            text=reason,
        )


def mark_account_as_compliant(account_id: UUID, reason: str):
    account = Account.objects.get(id=account_id)
    if account.delinquent:
        logger.info('mark-account-as-compliant', account_id=account_id, reason=reason)
        account.delinquent = False
        account.save()
        EventLog.objects.create(
            account_id=account_id,
            type=EventLog.NEW_COMPLIANT,
            text=reason,
        )


def charge_pending_invoices(account_id: UUID) -> Dict[str, int]:
    account = Account.objects.get(id=account_id)
    pending_invoices = account.invoices.payable().only('pk')
    logger.info('charge-pending-invoices', pending_invoices=pending_invoices)

    payment_transactions = []
    for invoice in pending_invoices:
        try:
            payment_transaction = invoices.pay_with_account_credit_cards(invoice.pk)
            if payment_transaction:
                payment_transactions.append(payment_transaction)
        except PreconditionError:
            continue

    reasons = get_reasons_account_is_violating_delinquent_criteria(account.id)
    if not reasons:
        mark_account_as_compliant(account.id, reason='Pending invoices have been paid')

    num_paid_invoices = len(payment_transactions)
    return {
        'num_paid_invoices': num_paid_invoices,
        'num_failed_invoices': len(pending_invoices) - num_paid_invoices
    }



from django.dispatch import receiver
from structlog import get_logger

from . import credit_card_registered
from ..models import CreditCard
from ..actions import accounts

logger = get_logger()


@receiver(credit_card_registered)
def credit_card_registered_handler(sender, credit_card: CreditCard, **kwargs):
    account = credit_card.account
    if not account.delinquent:
        return

    reasons = accounts.get_reasons_account_is_violating_delinquent_criteria(account.id)
    if not reasons:
        accounts.mark_account_as_compliant(
            account.id,
            reason='A valid credit card has been registered'
        )



from django.apps import AppConfig


class BillingConfig(AppConfig):
    name = 'billing'

    def ready(self):
        # noinspection PyUnresolvedReferences
        import billing.signals.handlers  # noqa: F401



# Generated by Django 2.2.12 on 2020-10-06 06:23

import uuid

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    dependencies = [
        ('billing', '0017_new_djmoney'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='delinquent',
            field=models.BooleanField(db_index=True, default=False),
        ),
        migrations.CreateModel(
            name='EventLog',
            fields=[
                ('id', models.UUIDField(
                    default=uuid.uuid4,
                    editable=False,
                    primary_key=True,
                    serialize=False
                )),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('type', models.CharField(
                    choices=[
                        ('NEW_DELINQUENT', 'New delinquent'),
                        ('NEW_COMPLIANT', 'New compliant')
                    ],
                    max_length=20
                )),
                ('text', models.CharField(blank=True, max_length=255)),
                ('account', models.ForeignKey(
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name='event_logs',
                    to='billing.Account'
                )),
            ],
            options={
                'verbose_name': 'event log',
                'verbose_name_plural': 'event logs',
            },
        ),
    ]



from datetime import date, datetime
from typing import Dict

from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.db.models import Count, Max, Prefetch, Q
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import NoReverseMatch, reverse
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext_lazy as _
from import_export import resources
from import_export.admin import ExportMixin
from import_export.fields import Field
from import_export.formats import base_formats
from moneyed.localization import format_money
from structlog import get_logger

from .actions import accounts, invoices
from .models import (
    Account, Charge, CreditCard, EventLog, Invoice, ProductProperty,
    Transaction,
)

logger = get_logger()


##############################################################
# Shared utilities

class AppendOnlyModelAdmin(admin.ModelAdmin):
    """
    Adapted from: https://gist.github.com/aaugustin/1388243
    """

    def get_readonly_fields(self, request, obj=None):
        # Make everything readonly (unless superuser)
        if request.user.is_superuser:
            return super().get_readonly_fields(request, obj)

        # We cannot call super().get_fields(request, obj) because that method calls
        # get_readonly_fields(request, obj), causing infinite recursion. Ditto for
        # super().get_form(request, obj). So we  assume the default ModelForm.

        f = self.fields or [f.name for f in self.model._meta.fields]
        f.extend(super().get_readonly_fields(request, obj))
        return f

    def has_change_permission(self, request, obj=None):
        # Allow viewing objects but not actually changing them (unless superuser)
        if request.user.is_superuser:
            return True
        return request.method in ['GET', 'HEAD'] and super().has_change_permission(request, obj)

    def get_actions(self, request):
        # Disable delete action (unless superuser)
        actions = super().get_actions(request)
        if not request.user.is_superuser:
            if 'delete_selected' in actions:
                del actions['delete_selected']
        return actions

    def has_delete_permission(self, request, obj=None):
        # Disable delete link on detail page (unless superuser)
        return request.user.is_superuser


account_owner_search_fields = ['account__owner__email', 'account__owner__first_name', 'account__owner__last_name']


def amount(obj):
    return format_money(obj.amount)


amount.admin_order_field = 'amount'  # type: ignore


def created_on(obj):
    return obj.created.date()


created_on.admin_order_field = 'created'  # type: ignore
created_on.short_description = 'created'  # type: ignore


def modified_on(obj):
    return obj.modified.date()


modified_on.admin_order_field = 'modified'  # type: ignore
modified_on.short_description = 'modified'  # type: ignore


def psp_admin_link(obj):
    if obj.psp_content_type is None or obj.psp_object_id is None:
        return '-'

    text = '{}: {}'.format(obj.psp_content_type.name, obj.psp_object_id)
    try:
        url = reverse(
            'admin:{}_{}_change'.format(
                obj.psp_content_type.app_label,
                obj.psp_content_type.model),
            args=[obj.psp_object_id])
        return format_html('<a href="{}">{}</a>', url, text)
    except NoReverseMatch:
        return None


psp_admin_link.short_description = 'PSP Object'  # type: ignore


def link_to_account(obj):
    account = obj.account
    text = str(account)
    url = reverse('admin:billing_account_change', args=(account.pk,))
    return format_html('<a href="{}">{}</a>', url, text)


link_to_account.admin_order_field = 'account'  # type: ignore
link_to_account.short_description = 'Account'  # type: ignore


def link_to_invoice(obj):
    invoice_id = obj.invoice_id
    if invoice_id is None:
        return '-'
    else:
        text = '#{}'.format(invoice_id)
        url = reverse('admin:billing_invoice_change', args=(invoice_id,))
        return format_html('<a href="{}">{}</a>', url, text)


link_to_invoice.admin_order_field = 'invoice'  # type: ignore
link_to_invoice.short_description = 'Invoice'  # type: ignore


##############################################################
# Credit Cards

class CreditCardValidFilter(admin.SimpleListFilter):
    title = _('Valid')
    parameter_name = 'valid'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        today = datetime.now().date()
        if self.value() == 'yes':
            return queryset.filter(expiry_date__gte=today)
        if self.value() == 'no':
            return queryset.filter(expiry_date__lt=today)


def credit_card_expiry(obj):
    return format_html('{}/{}', obj.expiry_month, obj.expiry_year)


credit_card_expiry.admin_order_field = 'expiry_date'  # type: ignore


def credit_card_is_valid(obj):
    return obj.is_valid()


credit_card_is_valid.boolean = True  # type: ignore

credit_card_is_valid.short_description = 'valid'  # type: ignore


@admin.register(CreditCard)
class CreditCardAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = ['number', created_on, link_to_account, 'type', 'status', credit_card_expiry,
                    credit_card_is_valid,
                    psp_admin_link]
    search_fields = ['number'] + account_owner_search_fields
    list_filter = ['type', 'status', CreditCardValidFilter]
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account']
    readonly_fields = ['created', 'modified', 'expiry_date']


class CreditCardInline(admin.TabularInline):
    model = CreditCard
    fields = readonly_fields = ['type', 'number', 'status', credit_card_expiry, created_on, psp_admin_link]
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


##############################################################
# Charges

def charge_deleted(charge):
    """
    We don't want to display a green check box to mean deleted.
    We use text instead to make sure there is no misinterpretation.
    """
    return 'Yes' if charge.deleted else 'No'


charge_deleted.short_description = 'Deleted'  # type: ignore

charge_deleted.admin_order_field = 'deleted'  # type: ignore


class ProductPropertyInline(admin.TabularInline):
    model = ProductProperty
    verbose_name_plural = 'Product props'
    ordering = ['name']


def product_properties(obj):
    return format_html_join(
        ',\n',
        '<strong>{}</strong>: {}',
        ((p.name, p.value) for p in obj.product_properties.all()))


product_properties.short_description = 'Product props'  # type: ignore


@admin.register(Charge)
class ChargeAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = ['type', created_on, charge_deleted, link_to_account, amount, 'product_code', product_properties,
                    'ad_hoc_label', link_to_invoice]
    search_fields = ['id', 'amount', 'amount_currency', 'product_code', 'product_properties__value', 'ad_hoc_label',
                     'invoice__id'] + account_owner_search_fields
    list_filter = ['deleted', 'amount_currency', 'product_code']
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account', 'invoice', 'reverses']
    readonly_fields = ['created', 'modified']
    inlines = [ProductPropertyInline]

    def get_queryset(self, request):
        # In this particular admin screen we want to see even the deleted charges
        qs = Charge.all_charges
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs \
            .prefetch_related('invoice') \
            .prefetch_related('product_properties')


class ChargeInline(admin.TabularInline):
    verbose_name_plural = 'Charges and Credits'
    model = Charge
    fields = readonly_fields = ['type', created_on, link_to_invoice, amount, 'product_code', product_properties,
                                'ad_hoc_label']
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related('product_properties')


#############################################################
# Transactions

class TransactionResource(resources.ModelResource):
    amount = Field()
    amount_currency = Field()

    class Meta:
        model = Transaction
        fields = ['id', 'created', 'modified', 'success', 'payment_method',
                  'credit_card_number', 'account__owner__email', 'invoice']

    def dehydrate_amount(self, tx):
        if tx.amount is not None:
            return tx.amount.amount

    def dehydrate_amount_currency(self, tx):
        if tx.amount is not None:
            return tx.amount.currency.code


@admin.register(Transaction)
class TransactionAdmin(ExportMixin, AppendOnlyModelAdmin):
    verbose_name_plural = 'Transactions'
    date_hierarchy = 'created'
    list_display = ['type', created_on, link_to_account, 'payment_method', 'credit_card_number', 'success',
                    link_to_invoice, amount, psp_admin_link]
    list_display_links = ['type']
    search_fields = ['credit_card_number', 'amount'] + account_owner_search_fields
    list_filter = ['payment_method', 'success', 'amount_currency']
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account', 'invoice']
    readonly_fields = ['created', 'modified']

    # Export
    resource_class = TransactionResource
    formats = (base_formats.CSV, base_formats.XLS, base_formats.JSON)  # Safe and useful formats.


class TransactionInline(admin.TabularInline):
    model = Transaction
    fields = readonly_fields = ['type', 'payment_method', 'credit_card_number', created_on, 'success', amount,
                                link_to_invoice, psp_admin_link]
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


##############################################################
# Invoices

class InvoiceResource(resources.ModelResource):
    """
    We separate the due amount into number and currency columns to make processing the exported data easier.
    Calculating the due amount is costly and we don't want to do it twice so we cache it.
    """
    due_amount = Field()
    due_amount_currency = Field()

    cc = Field()

    due_cache: Dict = {}

    class Meta:
        model = Invoice
        fields = ['id', 'account__owner__email', 'created', 'modified', 'due_date', 'status']

    def dehydrate_due_amount(self, invoice):
        due = self._due(invoice)
        if due is not None:
            return due.amount

    def dehydrate_due_amount_currency(self, invoice):
        due = self._due(invoice)
        if due is not None:
            return due.currency.code

    def _due(self, invoice):
        cached = self.due_cache.get(invoice.id)
        if cached:
            return cached
        due = self.calculate_due(invoice)
        self.due_cache[invoice.id] = due
        return due

    def calculate_due(self, invoice):
        due_total_monies = invoice.due().monies()
        if len(due_total_monies) == 1:
            due_total = due_total_monies[0]
            return due_total

    def dehydrate_cc(self, invoice):
        return invoice_account_cc(invoice)


class InvoiceDueFilter(admin.SimpleListFilter):
    title = _('Due')
    parameter_name = 'due'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        today = datetime.now().date()
        if self.value() == 'yes':
            return queryset.filter(due_date__lte=today)
        if self.value() == 'no':
            return queryset.filter(due_date__gt=today)


class InvoiceCCFilter(admin.SimpleListFilter):
    title = _('cc')
    parameter_name = 'cc'

    def lookups(self, request, model_admin):
        return (
            ('valid', _('Valid')),
            ('expired', _('Expired')),
            ('none', _('None')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        d = date.today()
        if self.value() == 'valid':
            return queryset.filter(account__credit_cards__expiry_date__gte=d)
        if self.value() == 'expired':
            return queryset.exclude(account__credit_cards__expiry_date__gte=d) \
                .exclude(account__credit_cards=None)
        if self.value() == 'none':
            return queryset.filter(account__credit_cards=None)


def assign_funds_to_invoice_button(obj):
    if obj.pk and obj.in_payable_state:
        return format_html(
            '<a href="{}">Assign existing funds to invoice</a>',
            reverse('admin:billing-assign-funds-to-invoice', args=[obj.pk]))
    else:
        return '-'


assign_funds_to_invoice_button.short_description = _('Assign funds')  # type: ignore


def do_assign_funds_to_invoice(request, invoice_id):
    accounts.assign_funds_to_invoice(invoice_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


def pay_invoice_with_cc_button(obj):
    if obj.pk and obj.in_payable_state:
        return format_html(
            '<a href="{}">Pay invoice with CC</a>',
            reverse('admin:billing-pay-invoice-with-cc', args=[obj.pk]))
    else:
        return '-'


pay_invoice_with_cc_button.short_description = _('Pay')  # type: ignore


def do_pay_invoice_with_cc(request, invoice_id):
    invoices.pay_with_account_credit_cards(invoice_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


def invoice_number(invoice):
    return str(invoice)


invoice_number.short_description = 'Invoice'  # type: ignore

invoice_number.admin_order_field = 'pk'  # type: ignore


def invoice_last_transaction(obj):
    dt = obj.last_transaction
    if dt:
        return dt.date()


invoice_last_transaction.short_description = 'Last transaction'  # type: ignore


def invoice_account_cc(obj):
    if obj.valid_credit_card_count > 0:
        return 'Valid'
    elif obj.credit_card_count > 0:
        return 'Expired'
    else:
        return 'None'


invoice_account_cc.short_description = 'cc'  # type: ignore


@admin.register(Invoice)
class InvoiceAdmin(ExportMixin, AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = [invoice_number, created_on, modified_on, link_to_account, invoice_account_cc,
                    'total_charges', 'due', 'due_date', invoice_last_transaction, 'status']
    list_filter = [InvoiceCCFilter, InvoiceDueFilter, 'status']
    search_fields = ['id', 'account__owner__email', 'account__owner__first_name', 'account__owner__last_name']
    ordering = ['-created']

    raw_id_fields = ['account']
    readonly_fields = ['created', 'modified', 'total_charges', 'due', assign_funds_to_invoice_button,
                       pay_invoice_with_cc_button]
    inlines = [ChargeInline, TransactionInline]

    # Export
    resource_class = InvoiceResource
    formats = (base_formats.CSV, base_formats.XLS, base_formats.JSON)  # Safe and useful formats.

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .select_related('account__owner') \
            .annotate(last_transaction=Max('transactions__created')) \
            .annotate(
            credit_card_count=Count('account__credit_cards'),
            valid_credit_card_count=Count('account__credit_cards',
                                          filter=Q(account__credit_cards__expiry_date__gte=date.today()))) \
            .only('id', 'created', 'modified', 'account__owner__email', 'account__owner__id', 'due_date', 'status')

    def get_urls(self):
        custom_urls = [
            url(r'^(?P<invoice_id>[0-9a-f-]+)/assign_funds_to_invoice/$',
                self.admin_site.admin_view(do_assign_funds_to_invoice),
                name='billing-assign-funds-to-invoice'),
            url(r'^(?P<invoice_id>[0-9a-f-]+)/pay/$',
                self.admin_site.admin_view(do_pay_invoice_with_cc),
                name='billing-pay-invoice-with-cc')
        ]
        return custom_urls + super().get_urls()


class InvoiceInline(admin.TabularInline):
    model = Invoice
    readonly_fields = [invoice_number, created_on, 'status', 'due_date', 'total_charges', 'due']
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


@admin.register(EventLog)
class EventLogAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    list_display = ('created', 'type', 'text', link_to_account)
    ordering = ('-created',)
    list_display_links = None

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related('account')

    def has_add_permission(self, request):
        return False


class EventLogInline(admin.TabularInline):
    model = EventLog
    readonly_fields = ('created', 'type', 'text')
    show_change_link = False
    can_delete = False
    extra = 0
    ordering = ('-created',)


##############################################################
# Accounts

class AccountCCFilter(admin.SimpleListFilter):
    title = _('cc')
    parameter_name = 'cc'

    def lookups(self, request, model_admin):
        return (
            ('valid', _('Valid')),
            ('expired', _('Expired')),
            ('none', _('None')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        d = date.today()
        if self.value() == 'valid':
            return queryset.filter(credit_cards__expiry_date__gte=d)
        if self.value() == 'expired':
            return queryset.exclude(credit_cards__expiry_date__gte=d) \
                .exclude(credit_cards=None)
        if self.value() == 'none':
            return queryset.filter(credit_cards=None)


def payable_invoice_count(obj):
    return len(obj.payable_invoice_ids)


def account_cc(obj):
    if obj.valid_credit_card_count > 0:
        return 'Valid'
    elif obj.credit_card_count > 0:
        return 'Expired'
    else:
        return 'None'


def create_invoices_button(obj):
    if obj.pk:
        return format_html(
            '<a class="button" href="{}">Create Invoices</a>',
            reverse('admin:billing-create-invoices', args=[obj.pk]),
        )
    else:
        return '-'


create_invoices_button.short_description = _('Create Invoices')  # type: ignore


class CreateInvoicesForm(forms.Form):
    due_date = forms.DateField()


def create_invoices_form(request, account_id):
    form = CreateInvoicesForm(request.POST or None, initial={'due_date': date.today()})
    if request.method == 'POST':
        if form.is_valid():
            accounts.create_invoices(
                account_id=account_id,
                due_date=form.cleaned_data['due_date'])
            # As confirmation take the user to the account overview.
            return HttpResponseRedirect(reverse('admin:billing_account_change', args=[account_id]))

    return render(
        request,
        'admin/billing/form.html',
        {
            'title': 'Create invoices',
            'form': form,
            'opts': Account._meta,  # Used to setup the navigation / breadcrumbs of the page
        }
    )


def do_create_invoices(request, account_id):
    accounts.create_invoices(account_id=account_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


def assign_funds_to_pending_invoices_button(obj):
    if obj.pk:
        return format_html(
            '<a class="button" href="{}">Assign existing funds to pending invoices</a>',
            reverse('admin:billing-assign-funds-to-pending-invoices', args=[obj.pk]),
        )
    else:
        return '-'


assign_funds_to_pending_invoices_button.short_description = _('Assign funds')  # type: ignore


def do_assign_funds_to_pending_invoices(request, account_id):
    accounts.assign_funds_to_account_pending_invoices(account_id=account_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


@admin.register(Account)
class AccountAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = [
        'owner', created_on, modified_on, payable_invoice_count, account_cc, 'currency',
        'status', 'delinquent'
    ]
    search_fields = ['id', 'owner__email', 'owner__first_name', 'owner__last_name']
    ordering = ['-created']
    list_filter = [AccountCCFilter, 'currency', 'status']
    list_select_related = True

    raw_id_fields = ['owner']
    readonly_fields = ['balance', 'created', 'modified', create_invoices_button,
                       assign_funds_to_pending_invoices_button]

    inlines = [
        CreditCardInline,
        ChargeInline,
        InvoiceInline,
        TransactionInline,
        EventLogInline
    ]

    def save_model(self, request, obj, form, change):
        if 'delinquent' in form.changed_data:
            if obj.delinquent:
                accounts.mark_account_as_delinquent(obj.id, reason='Manually')
            else:
                accounts.mark_account_as_compliant(obj.id, reason='Manually')
        super().save_model(request, obj, form, change)

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            url(
                r'^(?P<account_id>[0-9a-f-]+)/create_invoices/$',
                self.admin_site.admin_view(create_invoices_form),
                name='billing-create-invoices'
            ), url(
                r'^(?P<account_id>[0-9a-f-]+)/assign_funds_to_pending_invoices/$',
                self.admin_site.admin_view(do_assign_funds_to_pending_invoices),
                name='billing-assign-funds-to-pending-invoices'
            )
        ]
        return my_urls + urls

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .prefetch_related(Prefetch('invoices',
                                       queryset=Invoice.objects.payable().only('id'),
                                       to_attr='payable_invoice_ids')) \
            .annotate(
            credit_card_count=Count('credit_cards'),
            valid_credit_card_count=Count('credit_cards', filter=Q(credit_cards__expiry_date__gte=date.today())))



import logging
from collections import defaultdict

import progressbar
import structlog
from django.core.management.base import BaseCommand

from ...actions.accounts import assign_funds_to_account_pending_invoices
from ...models import Account


def set_debug(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


logger = structlog.get_logger()


class Command(BaseCommand):
    help = """For all accounts with pending invoices: tries to match unassigned funds to invoices.
              Pass -v 2 to see sql queries.
              """

    def add_arguments(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help='Counts the accounts with pending invoices and exit')
        parser.add_argument('--progress', action='store_true', help='Displays a progress bar')

    def handle(self, *args, **options):
        if options['verbosity'] >= 2:
            set_debug('django.db.backends')

        accounts = Account.objects.open().with_pending_invoices().only('id')

        dry_run = options['dry_run']

        logger.info('match-funds-start', dry_run=dry_run, accounts_with_pending_invoices=len(accounts))

        if dry_run:
            return

        if options['progress']:
            bar = progressbar.ProgressBar()
            accounts = bar(accounts)

        try:
            stats = defaultdict(lambda: 0)
            for account in accounts:
                try:
                    paid_invoices = assign_funds_to_account_pending_invoices(account_id=account.id)
                    stats_key = '{}_invoices'.format(len(paid_invoices))
                    stats[stats_key] += 1
                except Exception as ex:
                    logger.error('error', account_id=account.pk, ex=ex)
                    stats['error'] += 1
        finally:
            logger.info('match-funds-done', **stats)



import progressbar
import structlog
from django.core.management.base import BaseCommand

from ...actions.accounts import (
    get_accounts_which_delinquent_status_has_to_change,
    get_reasons_account_is_violating_delinquent_criteria,
    mark_account_as_compliant,
    mark_account_as_delinquent,
)
from ...models import Account

logger = structlog.get_logger()


class Command(BaseCommand):
    help = (
        'This command mark accounts as delinquent and vice versa when account has '
        'pending invoices or not valid credit cards registered'
    )

    def add_arguments(self, parser):
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help='Shows accounts which delinquent status is going to change'
        )
        parser.add_argument(
            '--progress',
            action='store_true',
            help='Displays a progress bar'
        )

    def handle(self, *args, **options):
        dry_run = options['dry_run']
        logger.info('update-delinquents-command', dry_run=dry_run)

        account_ids = Account.objects.values_list('id', flat=True)
        new_delinquent_account_ids, new_compliant_account_ids = (
            get_accounts_which_delinquent_status_has_to_change(account_ids)
        )

        logger.info(
            'update-delinquents-command',
            new_delinquent_accounts=len(new_delinquent_account_ids),
            new_compliant_accounts=len(new_compliant_account_ids),
        )
        if dry_run:
            return

        accounts = Account.objects.filter(id__in=new_delinquent_account_ids)
        if options['progress']:
            bar = progressbar.ProgressBar()
            accounts = bar(accounts)

        for account in accounts:
            reasons = get_reasons_account_is_violating_delinquent_criteria(account.id)
            mark_account_as_delinquent(account.id, reason='. '.join(reasons))

        accounts = Account.objects.filter(id__in=new_compliant_account_ids)
        if options['progress']:
            bar = progressbar.ProgressBar()
            accounts = bar(accounts)

        for account in accounts:
            mark_account_as_compliant(account.id, reason='Requirements met again')




from django.urls import path
from rest_framework import routers

from .views import (
    AccountView,
    CreditCardViewSet,
    pay_open_invoices_with_registered_credit_cards,
)

router = routers.SimpleRouter(trailing_slash=False)
router.register(
    r'account/credit-cards',
    CreditCardViewSet,
    basename='billing_creditcard'
)

urlpatterns = router.urls + [
    path('account', AccountView.as_view(), name='billing_account'),
    path(
        'pay/open/invoices/with/registered/credit/cards',
        pay_open_invoices_with_registered_credit_cards,
        name='pay_open_invoices_with_registered_credit_cards'
    ),
]



# Generated by Django 3.0.6 on 2020-05-22 15:04

from django.db import migrations
import djmoney.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0016_longer_product_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='charge',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, max_digits=12),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='amount',
            field=djmoney.models.fields.MoneyField(decimal_places=2, max_digits=12),
        ),
    ]



import logging

import structlog
from django.core.management.base import BaseCommand

from ...models import Invoice


def set_debug(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


logger = structlog.get_logger()


class Command(BaseCommand):
    help = """Displays all paid and cancelled invoices where due != 0"""

    def handle(self, *args, **options):
        if options['verbosity'] >= 2:
            set_debug('django.db.backends')

        invoices = Invoice.objects.exclude(status=Invoice.PENDING)

        logger.debug('audit-closed-invoices', non_pending_invoice_count=len(invoices))

        for invoice in invoices:
            due_total = invoice.due()
            due_monies = due_total.monies()
            if len(due_monies) != 1:
                logger.info('wrong-number-of-currencies', invoice_id=invoice.id, status=invoice.status,
                            currency_count=len(due_monies))
                continue
            due_value = due_monies[0]
            if due_value.amount != 0:
                logger.info('non-zero-due', invoice_id=invoice.id, status=invoice.status, due=due_value)
